@@ -26,16 +26,19 @@
 
 __all__ = ['namedlist', 'NO_DEFAULT', 'FACTORY']
 
-# all of this hassle with ast is solely to provide a decent __init__
-#  function, that takes all of the right arguments and defaults. but
+# All of this hassle with ast is solely to provide a decent __init__
+#  function, that takes all of the right arguments and defaults. But
 #  it's worth it to get all of the normal python error messages.
-# for other functions, like __repr__, we don't bother. __init__ is
-#  the only function where we really need the argument processing.
+# For other functions, like __repr__, we don't bother. __init__ is
+#  the only function where we really need the argument processing,
+#  because __init__ is the only function whose signature will vary
+#  per class.
 
 import ast as _ast
 import sys as _sys
 from keyword import iskeyword as _iskeyword
-from collections import Mapping as _Mapping
+import collections as _collections
+import abc as _abc
 
 _PY2 = _sys.version_info[0] == 2
 _PY3 = _sys.version_info[0] == 3
@@ -48,6 +51,8 @@ else:
 
 NO_DEFAULT = object()
 
+# Wrapper around a callable. Used to specify a factory function instead
+#  of a plain default value.
 class FACTORY(object):
     def __init__(self, callable):
         self._callable = callable
@@ -56,51 +61,77 @@ class FACTORY(object):
         return self._callable()
 
 
+########################################################################
 # Keep track of fields, both with and without defaults.
 class _Fields(object):
+    default_not_specified = object()
+
     def __init__(self, default):
         self.default = default
         self.with_defaults = []        # List of (field_name, default).
         self.without_defaults = []     # List of field_name.
 
-    def add_with_default(self, field_name, default):
-        if default is NO_DEFAULT:
-            self.add_without_default(field_name)
+    def add(self, field_name, default):
+        if default is self.default_not_specified:
+            if self.default is NO_DEFAULT:
+                # No default. There can't be any defaults already specified.
+                if len(self.with_defaults) != 0:
+                    raise ValueError('field {0} without a default follows fields '
+                                     'with defaults'.format(field_name))
+                self.without_defaults.append(field_name)
+            else:
+                self.add(field_name, self.default)
         else:
-            self.with_defaults.append((field_name, default))
+            if default is NO_DEFAULT:
+                self.add(field_name, self.default_not_specified)
+            else:
+                self.with_defaults.append((field_name, default))
 
-    def add_without_default(self, field_name):
-        if self.default is NO_DEFAULT:
-            # No default. There can't be any defaults already specified.
-            if self.with_defaults:
-                raise ValueError('field {0} without a default follows fields '
-                                 'with defaults'.format(field_name))
-            self.without_defaults.append(field_name)
+
+########################################################################
+# Validate and possibly sanitize the field and type names.
+class _NameChecker(object):
+    def __init__(self, typename):
+        self.seen_fields = set()
+        self._check_common(typename, 'Type')
+
+    def check_field_name(self, fieldname, rename, idx):
+        try:
+            self._check_common(fieldname, 'Field')
+            self._check_specific_to_fields(fieldname)
+        except ValueError as ex:
+            if rename:
+                return '_' + str(idx)
+            else:
+                raise
+
+        self.seen_fields.add(fieldname)
+        return fieldname
+
+    def _check_common(self, name, type_of_name):
+        # tests that are common to both field names and the type name
+        if len(name) == 0:
+            raise ValueError('{0} names cannot be zero '
+                             'length: {1!r}'.format(type_of_name, name))
+        if _PY2:
+            if not all(c.isalnum() or c=='_' for c in name):
+                raise ValueError('{0} names can only contain '
+                                 'alphanumeric characters and underscores: '
+                                 '{1!r}'.format(type_of_name, name))
+            if name[0].isdigit():
+                raise ValueError('{0} names cannot start with a '
+                                 'number: {1!r}'.format(type_of_name, name))
         else:
-            self.add_with_default(field_name, self.default)
+            if not name.isidentifier():
+                raise ValueError('{0} names names must be valid '
+                                 'identifiers: {1!r}'.format(type_of_name, name))
+        if _iskeyword(name):
+            raise ValueError('{0} names cannot be a keyword: '
+                             '{1!r}'.format(type_of_name, name))
 
-
-# Used for both the type name and field names. If is_type_name is
-#  False, seen_names must be provided. Raise ValueError if the name is
-#  bad.
-def _check_name(name, is_type_name=False, seen_names=None):
-    if len(name) == 0:
-        raise ValueError('Type names and field names cannot be zero '
-                         'length: {0!r}'.format(name))
-    if not all(c.isalnum() or c=='_' for c in name):
-        raise ValueError('Type names and field names can only contain '
-                         'alphanumeric characters and underscores: '
-                         '{0!r}'.format(name))
-    if _iskeyword(name):
-        raise ValueError('Type names and field names cannot be a keyword: '
-                         '{0!r}'.format(name))
-    if name[0].isdigit():
-        raise ValueError('Type names and field names cannot start with a '
-                         'number: {0!r}'.format(name))
-
-    if not is_type_name:
+    def _check_specific_to_fields(self, name):
         # these tests don't apply for the typename, just the fieldnames
-        if name in seen_names:
+        if name in self.seen_fields:
             raise ValueError('Encountered duplicate field name: '
                              '{0!r}'.format(name))
 
@@ -109,60 +140,54 @@ def _check_name(name, is_type_name=False, seen_names=None):
                              '{0!r}'.format(name))
 
 
-# Validate a field name. If it's a bad name, and if rename is True,
-#  then return a 'sanitized' name. Raise ValueError if the name is bad.
-def _check_field_name(name, seen_names, rename, idx):
-    try:
-        _check_name(name, seen_names=seen_names)
-    except ValueError as ex:
-        if rename:
-            return '_' + str(idx)
-        else:
-            raise
-
-    seen_names.add(name)
-    return name
-
-
 ########################################################################
-# member functions
+# Member functions for the generated class.
 
 def _repr(self):
-    return '{0}({1})'.format(self._name, ', '.join('{0}={1!r}'.format(name, getattr(self, name)) for name in self._fields))
-
+    return '{0}({1})'.format(self.__class__.__name__, ', '.join('{0}={1!r}'.format(name, getattr(self, name)) for name in self._fields))
 
 def _eq(self, other):
     return isinstance(other, self.__class__) and all(getattr(self, name) == getattr(other, name) for name in self._fields)
 
-
 def _ne(self, other):
     return not _eq(self, other)
 
-
 def _len(self):
     return len(self._fields)
 
-
 def _asdict(self):
     return {fieldname: getattr(self, fieldname) for fieldname in self._fields}
 
-
 def _getstate(self):
     return tuple(getattr(self, fieldname) for fieldname in self._fields)
 
-
 def _setstate(self, state):
     for fieldname, value in zip(self._fields, state):
         setattr(self, fieldname, value)
 
+def _getitem(self, idx):
+    return getattr(self, self._fields[idx])
+
+def _setitem(self, idx, value):
+    return setattr(self, self._fields[idx], value)
 
 def _iter(self):
     return (getattr(self, fieldname) for fieldname in self._fields)
-########################################################################
 
+def _count(self, value):
+    return sum(1 for v in iter(self) if v == value)
+
+def _index(self, value, start=NO_DEFAULT, stop=NO_DEFAULT):
+    # not the most efficient way to implement this, but it will work
+    l = list(self)
+    if start is NO_DEFAULT and stop is NO_DEFAULT:
+        return l.index(value)
+    if stop is NO_DEFAULT:
+        return l.index(value, start)
+    return l.index(value, start, stop)
 
 ########################################################################
-# the function that __init__ calls to do the actual work
+# The function that __init__ calls to do the actual work.
 
 def _init(self, *args):
     # sets all of the fields to their passed in values
@@ -173,12 +198,13 @@ def _init(self, *args):
             #  factory function: call it
             value = value()
         setattr(self, fieldname, value)
+
+
 ########################################################################
-
-
-# returns a function with name 'name', that calls another function 'chain_fn'
-# this is used to create the __init__ function with the right argument names and defaults, that
-#  calls into _init to do the real work
+# Returns a function with name 'name', that calls another function 'chain_fn'
+# This is used to create the __init__ function with the right argument names and defaults, that
+#  calls into _init to do the real work.
+# The new function takes args as arguments, with defaults as given.
 def _make_fn(name, chain_fn, args, defaults):
     args_with_self = ['self'] + list(args)
     arguments = [_ast.Name(id=arg, ctx=_ast.Load()) for arg in args_with_self]
@@ -211,15 +237,19 @@ def _make_fn(name, chain_fn, args, defaults):
     return locals_[name]
 
 
+########################################################################
+# The actual namedlist factory function. Needs a docstring.
 def namedlist(typename, field_names, default=NO_DEFAULT, rename=False,
               use_slots=True):
     # field_names must be a string or an iterable, consisting of fieldname
     #  strings or 2-tuples. Each 2-tuple is of the form (fieldname,
     #  default).
 
+    # Keeps track of the fields we're adding, with their defaults.
     fields = _Fields(default)
 
-    _check_name(typename, is_type_name=True)
+    # Validates field and type names.
+    name_checker = _NameChecker(typename)
 
     if isinstance(field_names, _basestring):
         # No per-field defaults. So it's like a namedtuple, but with
@@ -227,23 +257,18 @@ def namedlist(typename, field_names, default=NO_DEFAULT, rename=False,
         field_names = field_names.replace(',', ' ').split()
 
     # If field_names is a Mapping, change it to return the
-    #  (field_name, default) pairs, as if it were a list
-    if isinstance(field_names, _Mapping):
+    #  (field_name, default) pairs, as if it were a list.
+    if isinstance(field_names, _collections.Mapping):
         field_names = field_names.items()
 
-    # Parse and validate the field names.  Validation serves two
-    #  purposes: generating informative error messages and preventing
-    #  template injection attacks.
+    # Parse and validate the field names.
 
     # field_names is now an iterable. Walk through it,
     # sanitizing as needed, and add to fields.
 
-    seen_names = set()
     for idx, field_name in enumerate(field_names):
         if isinstance(field_name, _basestring):
-            field_name = _check_field_name(field_name, seen_names, rename,
-                                           idx)
-            fields.add_without_default(field_name)
+            default = fields.default_not_specified
         else:
             try:
                 if len(field_name) != 2:
@@ -254,36 +279,48 @@ def namedlist(typename, field_names, default=NO_DEFAULT, rename=False,
                 raise ValueError('field_name must be a 2-tuple: '
                                  '{0!r}'.format(field_name))
             default = field_name[1]
-            field_name = _check_field_name(field_name[0], seen_names, rename,
-                                           idx)
-            fields.add_with_default(field_name, default)
+            field_name = field_name[0]
+
+        # Okay: now we have the field_name and the default value (if any).
+        # Validate the name, and add the field.
+        fields.add(name_checker.check_field_name(field_name, rename, idx), default)
 
     all_field_names = tuple(fields.without_defaults + [name for name, default in
                                                        fields.with_defaults])
+    defaults = [default for _, default in fields.with_defaults]
 
-    __init = _make_fn('__init__', _init, all_field_names, [default for _, default in fields.with_defaults])
-
-    type_dict = {'__init__': __init,
+    type_dict = {'__init__': _make_fn('__init__', _init, all_field_names, defaults),
                  '__repr__': _repr,
                  '__eq__': _eq,
                  '__ne__': _ne,
                  '__len__': _len,
                  '__getstate__': _getstate,
                  '__setstate__': _setstate,
+                 '__getitem__': _getitem,
+                 '__setitem__': _setitem,
                  '__iter__': _iter,
+                 '__hash__': None,
+                 'count': _count,
+                 'index': _index,
                  '_asdict': _asdict,
-                 '_name': typename,
                  '_fields': all_field_names}
 
     if use_slots:
         type_dict['__slots__'] = all_field_names
 
-    # and finally, create and return the new type object
-    return type(typename, (object,), type_dict)
+    # Create the new type object.
+    t = type(typename, (object,), type_dict)
+
+    # Register its ABC's
+    _collections.Sequence.register(t)
+
+    # And return it.
+    return t
 
 
 if __name__ == '__main__':
     import unittest
+    import unicodedata
 
     # test both pickle and cPickle in 2.x, but just pickle in 3.x
     import pickle
@@ -331,6 +368,7 @@ if __name__ == '__main__':
             self.assertRaises(ValueError, namedlist, 'Point', '_field')
             self.assertRaises(ValueError, namedlist, 'Point', [('', 0)])
             self.assertRaises(ValueError, namedlist, '', 'x y')
+            self.assertRaises(ValueError, namedlist, 'Point', unicodedata.lookup('SUPERSCRIPT ONE'))
 
         def test_bad_defaults(self):
             # if specifying the defaults, must provide a 2-tuple
@@ -477,6 +515,9 @@ if __name__ == '__main__':
 
             self.assertEqual(list(iter(p)), [[1, 2, 3], 10, 20])
 
+            for expected, found in zip([[1, 2, 3], 10, 20], p):
+                self.assertEqual(expected, found)
+
         def test_fields(self):
             Point = namedlist('Point', 'x y z')
             self.assertEqual(Point._fields, ('x', 'y', 'z'))
@@ -529,6 +570,7 @@ if __name__ == '__main__':
             p = Point()
             # we should be able to create new attributes
             p.x = 3
+            self.assertEqual(p.x, 3)
 
         def test_rename(self):
             Point = namedlist('Point', ('abc', 'def'), rename=True)
@@ -576,7 +618,7 @@ if __name__ == '__main__':
             self.assertEqual(p.x, 0)
             self.assertEqual(p.y, 5)
 
-        def test_iterabale(self):
+        def test_iterable(self):
             Point = namedlist('Point', iter(['x', 'y']))
             p = Point(1, 2)
             self.assertEqual(p.x, 1)
@@ -614,6 +656,55 @@ if __name__ == '__main__':
             b = A()
             self.assertEqual(b.x, [])
 
+        def test_unhashable(self):
+            Point = namedlist('Point', 'a b')
+            p = Point(1, 2)
+            self.assertRaises(TypeError, hash, p)
+
+        def test_getitem(self):
+            Point = namedlist('Point', 'a b')
+            p = Point(1, 2)
+            self.assertEqual((p[0], p[1]), (1, 2))
+            self.assertEqual(list(p), [1, 2])
+            self.assertRaises(IndexError, p.__getitem__, 2)
+
+        def test_setitem(self):
+            Point = namedlist('Point', 'a b')
+            p = Point(1, 2)
+            p[0] = 10
+            self.assertEqual(list(p), [10, 2])
+            p[1] = 20
+            self.assertEqual(list(p), [10, 20])
+            self.assertRaises(IndexError, p.__setitem__, 2, 3)
+
+        def test_container(self):
+            # I'm not sure there's much sense in this, but list is a container
+            Point = namedlist('Point', 'a b')
+            p = Point(1, 2)
+            self.assertIn(2, p)
+
+        def test_ABC(self):
+            Point = namedlist('Point', 'a b c')
+            p = Point(1, 2, 2)
+            self.assertIsInstance(p, _collections.Container)
+            self.assertIsInstance(p, _collections.Iterable)
+            self.assertIsInstance(p, _collections.Sized)
+            self.assertIsInstance(p, _collections.Sequence)
+
+            self.assertEqual(list(reversed(p)), [2, 2, 1])
+            self.assertEqual(p.count(0), 0)
+            self.assertEqual(p.count(2), 2)
+            self.assertRaises(ValueError, p.index, 0)
+            self.assertEqual(p.index(2), 1)
+            self.assertEqual(p.index(1), 0)
+
+            A = namedlist('A', 'a b c d e f g h i j')
+            a = A(0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
+            self.assertEqual(a.index(0, 1), 1)
+            self.assertEqual(a.index(0, 5), 5)
+            self.assertEqual(a.index(0, 1, 3), 1)
+            self.assertEqual(a.index(0, 5, 12), 5)
+            self.assertRaises(ValueError, a.index, 0, 12)
 
 
     unittest.main()
